require 'iowa/ImageSize'
require 'iowa/Webcache'

# Notes on tags and tag logic:
# Currently all "builtin" tags are defined in two files, DynamicElements.rb
# and Form.rb.  This should be broken up and restructured.  Tags, or at least
# familes of tags (i.e. all input type tags are a family) should each be
# be defined in a separate file, and there should be some sort of "include path"
# for finding tags.  This makes it easy to add new tags without editing any
# core files.

module Iowa

# Subclass of Tag that defines an Iowa anchor element.

	class A < Tag

		# The default binding for an anchor is to establish an action
		# with a literal association.  i.e. oid="frumption" binds a
		# call to "frumption" as the action for the element.
	
		def defaultBindings
			@bindings[Caction] = LiteralAssociation.new(@name)
		end
		
		def handleRequest(context)
			if(context.actionID == context.elementID)
				context.setAction(@bindings[Caction])
				context.element_attributes = @attributes.dup
			end
			
			handleChildren(:handleRequest, context)
		end
		
		def handleResponse(context)
			context.response << openTag(Chref, context.actionURL, context)
			handleChildren(:handleResponse, context)
			context.response << closeTag
		end
	end
	
	class Img < Tag

		def webcache_size; 100; end
		
		def initialize(*args)
			@@image_size_cache ||= {}
			@@webcache = Iowa::Webcache.new(webcache_size)
			super(*args)
		end
		
		def defaultBindings
			@bindings[Caction] = LiteralAssociation.new(@name)
		end

		def get_image_size_by_path(path)
			File.open(path) {|fh| Iowa::ImageSize.new(fh)}
		end

		def get_image_size_by_data(data)
			Iowa::ImageSize.new(data.to_s)
		end

		# Change this so it can take a local path OR a URL.
		def handleResponse(context)
			image_url,image_path = context.getBinding(@bindings[Caction])
			img = nil
			if image_path.to_s != '' && FileTest.exists?(image_path)
				img = nil
				unless @@image_size_cache.has_key?(Cimage_path)
					img = get_image_size_by_path(image_path)
					@@image_size_cache[image_path] = [img,File.stat(image_path)]
				else
					img,st = @@image_size_cache[image_path]
					if File.stat(image_path).mtime != st.mtime
						img = get_image_size_by_path(image_path)
						@@image_size_cache[image_path] = [img.File.stat(image_path)]
					end
				end
			elsif image_url.to_s != ''
				if data = @@webcache[image_url]
					@@image_size_cache[image_url] = get_image_size_by_data(data)
				end
				img = @@image_size_cache[image_url]
			end
			if img
				if (img.height) and (img.width)
					context.response << openTag(Csrc,image_url,Cheight,img.height,Cwidth,img.width,context)
				else
					context.response << openTag(Csrc,image_url,context)
				end
				handleChildren(:handleResponse, context)
				context.response << closeTag
			else
				context.response << openTag(Csrc,image_url,context)
				handleChildren(:handleResponse, context)
				context.response << closeTag
			end
		end
	end
				
			
	# Subclass of Tag that defines an Iowa string element.
	# Note that Strings are usually referred to in a template
	# via the shortcut syntax of '@foo'.
	
	class DynamicString < Tag
	
		def defaultBindings
			@bindings[Cvalue] = PathAssociation.new(@name)
		end

		# Short circuits calls back to Element#handleRequest and thus
		# Element#handleChildren since a DynamicString will never have children.

		def handleRequest(context)
		end

		def handleResponse(context)
			val = context.getBinding(@bindings[Cvalue])
			if(val)
				context.response << "#{val}"
			end
		end
	end
	
	# Subclass of Tag that implements an Iowa element for repeating
	# portions of a template.
	
	class Repeat < Tag

		def defaultBindings
			@bindings[Clist] = nil
			@bindings[Citem] = nil
			@bindings[Ckey] = nil
		end
		
		def handleRequestOrResponse(method, context)
			if @bindings[Clist].nil? and @bindings[Citem]
				@bindings[Clist] = PathAssociation.new(@name)
			elsif @bindings[Citem].nil? and @bindings[Clist]
				@bindings[Citem] = PathAssociation.new(@name)
			elsif @bindings[Clist].nil? and @bindings[Citem].nil?
				b = PathAssociation.new(@name + C_list)
				if context.testBinding(b)
					@bindings[Clist] = b
					@bindings[Citem] = PathAssociation.new(@name)
				else
					@bindings[Clist] = PathAssociation.new(@name)
					@bindings[Citem] = PathAssociation.new(C_item)
				end
			end

			# Now, check to see if the binding for item responds.  If it does not,
			# create it.

			context.createAccessor(@bindings[Citem]) unless context.testBinding(@bindings[Citem])

			list = context.getBinding(@bindings[Clist])
			key = @bindings[Ckey]
			if list
				context.pushElement
				for item in list
					context.setBindingNow(@bindings[Citem], item)
					if(key)
						context.elementID = context.getBinding(key)
					else
						context.nextElement!
					end
					handleChildren(method, context)
				end
				context.popElement
			end
		end
	end
	
	# Subclass of Repeat.  Syntactic sugar.
	
	class Table < Repeat	
		def handleResponse(context)
			context.response << openTag(context)
			super(context)
			context.response << closeTag
		end
	end
	
	# Subclass of Repeat.  Syntactic sugar.
	
	class UL < Repeat	
		def handleResponse(context)
			context.response << openTag(context)
			super(context)
			context.response << closeTag
		end
	end

	# Subclass of Repeat.  Syntactic sugar.
	
	class OL < Repeat	
		def handleResponse(context)
			context.response << openTag(context)
			super(context)
			context.response << closeTag
		end
	end

	# Subclass of Repeat.  Syntactic sugar.
	
	class TR < Repeat		
		def handleResponse(context)
			context.response << openTag(context)
			super(context)
			context.response << closeTag
		end
	end

	class SideEffect < Tag

		def handleResponse(context)
			context.response << openTag(context)
			super(context)
			context.response << closeTag
		end
		
		def handleRequest(context)
			handleChildren(:handleRequest, context)
		end
	end

	
	# Subclass of Tag that allows for optional sections within a template.
	# Evaluation is simple.  If the condition (the item referenced by the
	# oid attribute in the tag) evaulates to true (Ruby's notion of true)
	# then the body of the <IF> tag will be displayed.  Otherwise it will
	# not be displayed.
	
	class If < Tag

		def defaultBindings
			@bindings[Ccondition] = PathAssociation.new(@name)
		end
		
		def testCondition(condition)
			condition
		end
	
		def handleRequestOrResponse(method, context)
			condition = context.getBinding(@bindings[Ccondition])
			context.element_attributes = @attributes
			if testCondition(condition)
				handleChildren(method, context)
			end
		end
	end
	
	# Subclass of If that implements inverse functionality.  If the
	# condition evaluates to false, then the body of the <UNLESS> tag
	# will be displayed.
	
	class Unless < If
		def testCondition(condition)
			!condition
		end
	end
	
	#Deprecated; will be removed.
	# Can we have an Else, too?

	class Span < If
	end
	
	class IfCondition < If
		def handleResponse(context)
			context.response << openTag(context)
			super(context)
			context.response << closeTag
		end
	end

#	class TextElement < Element
#	
#		def initialize(name)
#			super(name, {}, {})
#		end
#
#		def handleRequest(context)
#		end
#
#		def handleResponse(context)
#			context.response << @name
#		end
#	end

	class TextElement < String
		def handleRequest(context); end
		def handleResponse(context); context.response << self; end
	end
	
	class BodyContent < Element

		def handleRequestOrResponse(method, context)
			root = context.popRoot
			root.handleChildren(method, context)
			context.pushRoot root
		end
	end

end
