require 'thread'

module Iowa

	# Simple LRU cache for caching pages.

	class LRUCache

		# Initialize the empty LRU cache, and set the maximum number of elements
		# that it will contain or the max time to live (in seconds) of a cache item.
		# If both a max and a maxttl are set, the max is a hard maximum that will
		# not be exceeded.
		def initialize(max=20, maxttl=nil)
			@cache = {}
			@queue = []
			@ttl_cache = {}
			@max = max
			@maxttl = maxttl
			@finalizers = []
			@mutex = Mutex.new
		end

		# Check to see if the cache contains the given key.

		def include?(key)
			@cache.include?(key)
		end

		# Return the element identified by the given key.

		def [](key)
			r = nil
			@mutex.synchronize do
				if @cache.has_key? key
					r = @cache[key]
					@queue.push @queue.delete(key)
					@ttl_cache[key] = Time.now.to_i
				end
			end
			r
		end
		
		def prune
			obj = @queue.shift
			do_finalization(obj,@cache[obj])
			@cache.delete obj
			@ttl_cache.delete obj
		end

		def attach(key, val)
			@cache[key] = val
			@ttl_cache[key] = Time.now.to_i
			@queue.push key
		end
		
		# Set the element of the cache identified by the given key.  

		def []=(key, val)
			@mutex.synchronize do
				prune if @queue.length == @max
				if @maxttl
					expiration = Time.now.to_i - @maxttl
					prune while !@queue.empty? && (@ttl_cache[@queue.first].to_i < expiration)
				end
				attach(key, val) if @queue.length < @max
			end
			@queue
		end

		# Allows one to set the maximum size of the cache queue.  If the queue
		# is currently larger than the size that it is being set to, elements
		# will be expired until the queue is at the maximum size.

		def size=(max)
			@mutex.synchronize do
				prune while @queue.length > max
				@max = max
			end
		end
		alias maxsize= size=

		# Return the maximum size of the cache.
		def maxsize
			@max
		end

		# Return the current size of the cache.
		def size
			@queue.length
		end

		# Return a copy of current set of keys to cache elements.
		def queue
			@queue.dup
		end
		
		# Adds a finalization method to the cache that will be called before
		# the object in the cache is expired.	
		def add_finalizer(*args,&block)
			@finalizers.push [block,args]
		end
		
		# Is called when an object in the cache is expired.  Iterates through
		# the defined finalization methods, if any.
		def do_finalization(key,obj)
				begin
					@finalizers.each do |f|
						f[0].call(key,obj,*f[1])
					end
				rescue Exception => e
					puts e, e.backtrace
					raise e
				end
		end
	end
end
